#!/bin/bash

BORG_DIR=${BORG_DIR:-/opt/borg}
BACKUP_HOST=${BACKUP_HOST:-backup.jim.sh}
BACKUP_USER=${BACKUP_USER:-jim-backups}
BACKUP_REPO=${BACKUP_REPO:-borg/$(hostname)}

# Use stable host ID in case MAC address changes
HOSTID="$(hostname -f)@$(python -c 'import uuid;print(uuid.getnode())')"

function error_handler() {
    echo "Error at $1 line $2:"
    echo -n '>>> ' ; tail -n +"$2" < "$1" | head -1
    echo "... exited with code $3"
    exit "$3"
}
trap 'error_handler ${BASH_SOURCE} ${LINENO} $?' ERR
set -o errexit
set -o errtrace

if [ -e "$BORG_DIR" ]; then
   echo "Error: BORG_DIR $BORG_DIR already exists; giving up"
   exit 1
fi

# Make a temp dir to work in
mkdir "$BORG_DIR"
TMP=$(mktemp -d --tmpdir="$BORG_DIR")

# Install some cleanup handlers
cleanup()
{
    set +o errexit
    set +o errtrace
    trap - ERR
    ssh -o ControlPath="$TMP"/ssh-control -O exit x >/dev/null 2>&1
    rm -rf -- "$TMP"
}
cleanup_int()
{
    echo
    cleanup
    exit 1
}
trap cleanup 0
trap cleanup_int 1 2 15

msg()
{
    color="$1"
    shift
    echo -ne "\033[1;${color}m===\033[0;${color}m" "$@"
    echo -e "\033[0m"
}
log(){ msg 33 "$@" ; }
notice() { msg 32 "$@" ; }
warn() { msg 31 "$@" ; }
error() { msg 31 "Error:" "$@" ; exit 1 ; }

# Install required packages
install_dependencies()
{
    NEED=
    check() {
        command -v "$1" >/dev/null || NEED+=" $2"
    }
    check borg borgbackup
    if [ -n "${NEED:+x}" ]; then
        log "Need to install packages: $NEED"
        apt install --no-upgrade "$NEED"
    fi
}

# Create wrapper to execute borg
create_borg_wrapper()
{
    BORG=${BORG_DIR}/borg.sh
    BORG_REPO="ssh://${BACKUP_USER}@${BACKUP_HOST}/./${BACKUP_REPO}"

    cat >"$BORG" <<EOF
#!/bin/sh

export BORG_REPO=${BORG_REPO}
export BORG_PASSCOMMAND=${BORG_DIR}/print-passphrase
export BORG_HOST_ID=${HOSTID}
export BORG_BASE_DIR=${BORG_DIR}
export BORG_CACHE_DIR=${BORG_DIR}/cache
export BORG_CONFIG_DIR=${BORG_DIR}/config
export BORG_RSH="ssh -F ${BORG_DIR}/ssh/config"
exec borg "\$@"
EOF
    chmod +x "$BORG"
    if ! "$BORG" -h >/dev/null ; then
        error "Can't run the new borg wrapper; does borg work?"
    fi

}

print_random_key()
{
    dd if=/dev/urandom | tr -dc 'a-zA-Z0-9' | head -c 16
}

generate_keys()
{
    PASS_SSH=$(print_random_key)
    PASS_REPOKEY=$(print_random_key)
    cat >"${BORG_DIR}/print-passphrase" <<EOF
#!/bin/sh
echo $PASS_REPOKEY
EOF
    chmod 700 "${BORG_DIR}/print-passphrase"
}

# Run a command on the remote host over an existing SSH tunnel
run_ssh_command()
{
    ssh -o ControlPath="$TMP"/ssh-control use-existing-control-tunnel "$@"
}

# Configure SSH key-based login
configure_ssh()
{
    SSH=$BORG_DIR/ssh
    mkdir "$SSH"

    # Create keys
    log "Creating SSH keys"
    ssh-keygen -N "" -t ecdsa \
               -C "backup-appendonly@$HOSTID" -f "$SSH/id_ecdsa_appendonly"
    ssh-keygen -N "$PASS_SSH" -t ecdsa \
               -C "backup@$HOSTID" -f "$SSH/id_ecdsa"

    # Create config snippets
    log "Creating SSH config and wrapper script"
    cat >> "$SSH/config" <<EOF
User $BACKUP_USER
ControlPath none
ServerAliveInterval 120
Compression no
UserKnownHostsFile $SSH/known_hosts
ForwardX11 no
ForwardAgent no
BatchMode yes
IdentityFile $SSH/id_ecdsa_appendonly

Host backup-appendonly
    HostName $BACKUP_HOST
    IdentityFile $SSH/id_ecdsa_appendonly

Host backup
    HostName $BACKUP_HOST
    IdentityFile $SSH/id_ecdsa
EOF

    # Connect to backup host, using persistent control socket
    log "Connecting to server"
    log "Please enter password; look in Bitwarden for: ${BACKUP_USER}@${BACKUP_HOST}"
    ssh -F "$SSH/config" -o BatchMode=no -o PubkeyAuthentication=no \
        -o ControlMaster=yes -o ControlPath="$TMP/ssh-control" \
        -o StrictHostKeyChecking=accept-new \
        -f backup sleep 600
    if ! run_ssh_command true >/dev/null 2>&1 </dev/null ; then
        error "SSH failed"
    fi
    log "Connected to ${BACKUP_USER}@${BACKUP_HOST}"

    # Since we now have an SSH connection, check that the repo doesn't exist
    if run_ssh_command "test -e $BACKUP_REPO" ; then
        error "$BACKUP_REPO already exists on the server, bailing out"
    fi

    # Copy SSH keys to the server's authorized_keys file, removing any
    # existing keys with this HOSTID.
    log "Setting up SSH keys on remote host"
    cmd="borg serve --restrict-to-repository ~/$BACKUP_REPO"

    keys=".ssh/authorized_keys"
    backup="${keys}.old-$(date +%Y%m%d-%H%M%S)"
    run_ssh_command "mkdir -p .ssh; chmod 700 .ssh; touch $keys"
    run_ssh_command "mv $keys $backup; sed '/@$HOSTID\$/d' < $backup > $keys"
    run_ssh_command "if cmp -s $backup $keys; then rm $backup ; fi"
    run_ssh_command "cat >> .ssh/authorized_keys" <<EOF
command="$cmd --append-only",restrict $(cat "$SSH/id_ecdsa_appendonly.pub")
command="$cmd",restrict $(cat "$SSH/id_ecdsa.pub")
EOF

    # Test that everything worked
    log "Testing SSH login with new key"
    if ! ssh -F "$SSH/config" -T backup-appendonly borg --version </dev/null ; then
        error "Logging in with a key failed -- is server set up correctly?"
    fi
    log "Remote connection OK!"
}

# Create the repository on the server
create_repo()
{
    log "Creating repo $BACKUP_REPO"
    # Create repo
    $BORG init --make-parent-dirs --encryption repokey
}

# Export keys as HTML page
export_keys()
{
    log "Exporting keys"
    $BORG key export --paper '' "${BORG_DIR}/key.txt"
    chmod 600 "${BORG_DIR}/key.txt"
    cat >>"${BORG_DIR}/key.txt" <<EOF

Repository: ${BORG_REPO}
Passphrase: ${PASS_REPOKEY}
EOF
}

# Create helper scripts to backup, prune, and mount
create_scripts()
{
    cat > "${BORG_DIR}/backup.sh" <<EOF
#!/bin/bash

BORG=$BORG_DIR/borg.sh
set -e

# Explicitly list a bunch of directories to back up, in case they come
# from different filesystems.  If not, duplicates have no effect.
DIRS="/"
for DIR in /usr /var /home /boot /efi ; do
        if [ -e $DIR ] ; then
                DIRS="\$DIRS \$DIR"
        fi
done

# Allow dirs to be overridden
BORG_BACKUP_DIRS=\${BORG_BACKUP_DIRS:-\$DIRS}

echo "Backing up: $DIRS"

\$BORG create \\
        --verbose \\
        --list \\
        --filter E \\
        --stats \\
        --exclude-caches \\
        --one-file-system \\
        --checkpoint-interval 300 \\
        --compression zstd,3 \\
        ::'{hostname}-{now:%Y%m%d-%H%M%S}' \\
        \$BORG_BACKUP_DIRS

\$BORG check \\
        --verbose \\
        --last 10
EOF

    cat > "${BORG_DIR}/prune.sh" <<EOF
#!/bin/bash

BORG=$BORG_DIR/borg.sh
set -e

echo "=== Need SSH key passphrase.  Check Bitwarden for:"
echo "===   borg $(hostname)"
echo "===   read-write SSH key"
\$BORG prune \\
        --rsh="ssh -F ${BORG_DIR}/ssh/config -i ${SSH}/id_ecdsa"
        --verbose \\
        --stats \\
        --keep-within=7d \\
        --keep-daily=14 \\
        --keep-weekly=8 \\
        --keep-monthly=-1
EOF

    cat > "${BORG_DIR}/mount.sh" <<EOF
#!/bin/bash

BORG=$BORG_DIR/borg.sh
set -e

if [ -z "\$1" ] ; then
        echo "Usage: \$0 <mountpoint>"
        exit 1
fi

\$BORG mount \\
        :: \$1

echo "Unmount with: fusermount -u \$1"

EOF
}

configure_systemd()
{
    TIMER=borg-backup.timer
    SERVICE=borg-backup.service
    TIMER_UNIT=${BORG_DIR}/${TIMER}
    SERVICE_UNIT=${BORG_DIR}/${SERVICE}

    log "Creating systemd files"

    cat > "$TIMER_UNIT" <<EOF
[Unit]
Description=Borg backup to ${BACKUP_HOST}

[Timer]
OnCalendar=*-*-* 01:00:00
RandomizedDelaySec=1800
FixedRandomDelay=true
Persistent=true

[Install]
WantedBy=timers.target
EOF

    cat >> "$SERVICE_UNIT" <<EOF
[Unit]
Description=Borg backup to ${BACKUP_HOST}

[Service]
Type=simple
ExecStart=${BORG_DIR}/backup.sh
Nice=10
IOSchedulingClass=best-effort
IOSchedulingPriority=6
EOF

    log "Setting up systemd"
    if (
        ln -sfv "${TIMER_UNIT}" /etc/systemd/system &&
        ln -sfv "${SERVICE_UNIT}" /etc/systemd/system &&
        systemctl --no-ask-password daemon-reload &&
        systemctl --no-ask-password enable ${TIMER} &&
        systemctl --no-ask-password start ${TIMER}
    ); then
        log "Backup timer installed:"
        systemctl list-timers ${TIMER}
    else
        warn ""
        warn "Systemd setup failed"
        warn "Do something like this to configure automatic backups:"
        echo "     sudo ln -sfv \"${TIMER_UNIT}\" /etc/systemd/system &&"
        echo "     sudo ln -sfv \"${SERVICE_UNIT}\" /etc/systemd/system &&"
        echo "     sudo systemctl daemon-reload &&"
        echo "     sudo systemctl enable ${TIMER} &&"
        echo "     sudo systemctl start ${TIMER}"
        warn ""
    fi
}

log "Configuration:"
log "  Backup server host: ${BACKUP_HOST}"
log "  Backup server user: ${BACKUP_USER}"
log "     Repository path: ${BACKUP_REPO}"

install_dependencies
create_borg_wrapper
generate_keys
configure_ssh
create_repo
export_keys
create_scripts
configure_systemd

echo
notice "Add these two passwords to Bitwarden:"
notice ""
notice "      Name: borg $(hostname)"
notice "  Username: repo key"
notice "  Password: $PASS_REPOKEY"
notice ""
notice "      Name: borg $(hostname)"
notice "  Username: read-write ssh key"
notice "  Password: $PASS_SSH"
notice ""
notice "You should also print out the full repo key: ${BORG_DIR}/key.txt"
notice ""
notice "To run a manual backup:"
notice "  sudo systemctl start borg-backup"
notice ""
notice "and to see logs:"
notice "  journalctl -u borg-backup"
echo

echo "All done"
